package game.window;

import static org.lwjgl.glfw.Callbacks.glfwFreeCallbacks;
import static org.lwjgl.glfw.GLFW.*;
import static org.lwjgl.system.MemoryUtil.NULL;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.lwjgl.BufferUtils;
import org.lwjgl.Version;
import org.lwjgl.glfw.GLFW;
import org.lwjgl.glfw.GLFWCharCallback;
import org.lwjgl.glfw.GLFWCursorPosCallback;
import org.lwjgl.glfw.GLFWErrorCallback;
import org.lwjgl.glfw.GLFWFramebufferSizeCallback;
import org.lwjgl.glfw.GLFWImage;
import org.lwjgl.glfw.GLFWKeyCallback;
import org.lwjgl.glfw.GLFWMouseButtonCallback;
import org.lwjgl.glfw.GLFWScrollCallback;
import org.lwjgl.glfw.GLFWVidMode;
import org.lwjgl.glfw.GLFWWindowCloseCallback;
import org.lwjgl.glfw.GLFWWindowFocusCallback;
import org.lwjgl.glfw.GLFWWindowPosCallback;
import org.lwjgl.glfw.GLFWWindowRefreshCallback;
import org.lwjgl.opengl.GL;
import org.lwjgl.system.APIUtil;

import game.log.Log;

public abstract class Window {
	public static long getTime() {
		return glfwGetTimerValue() / (glfwGetTimerFrequency() / 1000000L);
	}
	public static void setWindowed(int xpos, int ypos, int xsize, int ysize) {
		if(window != NULL)
			glfwSetWindowMonitor(window, NULL, xpos, ypos, xsize, ysize, GLFW_DONT_CARE);
	}
	public static void setFullscreen(int width, int height, int refresh) {
		if(window != NULL) {
			long monitor = glfwGetPrimaryMonitor();
			if(monitor != NULL)
				glfwSetWindowMonitor(window, monitor, 0, 0, width, height, refresh);
		}
	}
	public static boolean getKey(Keysym code) {
		return window != NULL && glfwGetKey(window, CODES[code.ordinal()]) == GLFW_PRESS;
	}
	public static void setTitle(String title) {
		if(window != NULL)
			glfwSetWindowTitle(window, title);
	}
	public static void setIcon(byte[] icon, int w, int h) {
		if(window != NULL) {
			GLFWImage img = GLFWImage.create();
			img.width(w);
			img.height(h);
//			java.nio.ByteBuffer pbuf = ;
			img.pixels(BufferUtils.createByteBuffer(icon.length).put(icon));
//			img.pixels(icon.length / 4);
			GLFWImage.Buffer buf = GLFWImage.create(1).put(img);
			glfwSetWindowIcon(window, buf);
		}
	}
	public static String getClipboard() {
		return window == NULL ? null : glfwGetClipboardString(window);
	}
	public static void setClipboard(String text) {
		if(window != NULL)
			glfwSetClipboardString(window, text);
	}
	public static void swapBuffers() {
		if(window != NULL)
			glfwSwapBuffers(window);
	}
	public static void grabCursor(boolean grab) {
		if(window != NULL)
			glfwSetInputMode(window, GLFW_CURSOR, grab ? GLFW_CURSOR_DISABLED : GLFW_CURSOR_NORMAL);
	}
	public static void setVSync(boolean sync) {
		if(window != NULL)
			glfwSwapInterval(sync ? 1 : 0);
	}
	
	private static final int[] CODES = new int[Keysym.values().length];
	private static final Keysym[] KEYS = new Keysym[GLFW_KEY_LAST];
	private static final List<WindowEvent> EVENTS = new ArrayList<WindowEvent>();
	
	private static long window;
	
	private static void addKey(Keysym key, int code) {
		CODES[key.ordinal()] = code;
		KEYS[code - 1] = key;
	}
	
	static {
		addKey(Keysym.N0, GLFW_KEY_0);
		addKey(Keysym.N1, GLFW_KEY_1);
		addKey(Keysym.N2, GLFW_KEY_2);
		addKey(Keysym.N3, GLFW_KEY_3);
		addKey(Keysym.N4, GLFW_KEY_4);
		addKey(Keysym.N5, GLFW_KEY_5);
		addKey(Keysym.N6, GLFW_KEY_6);
		addKey(Keysym.N7, GLFW_KEY_7);
		addKey(Keysym.N8, GLFW_KEY_8);
		addKey(Keysym.N9, GLFW_KEY_9);
		
		addKey(Keysym.A, GLFW_KEY_A);
		addKey(Keysym.B, GLFW_KEY_B);
		addKey(Keysym.C, GLFW_KEY_C);
		addKey(Keysym.D, GLFW_KEY_D);
		addKey(Keysym.E, GLFW_KEY_E);
		addKey(Keysym.F, GLFW_KEY_F);
		addKey(Keysym.G, GLFW_KEY_G);
		addKey(Keysym.H, GLFW_KEY_H);
		addKey(Keysym.I, GLFW_KEY_I);
		addKey(Keysym.J, GLFW_KEY_J);
		addKey(Keysym.K, GLFW_KEY_K);
		addKey(Keysym.L, GLFW_KEY_L);
		addKey(Keysym.M, GLFW_KEY_M);
		addKey(Keysym.N, GLFW_KEY_N);
		addKey(Keysym.O, GLFW_KEY_O);
		addKey(Keysym.P, GLFW_KEY_P);
		addKey(Keysym.Q, GLFW_KEY_Q);
		addKey(Keysym.R, GLFW_KEY_R);
		addKey(Keysym.S, GLFW_KEY_S);
		addKey(Keysym.T, GLFW_KEY_T);
		addKey(Keysym.U, GLFW_KEY_U);
		addKey(Keysym.V, GLFW_KEY_V);
		addKey(Keysym.W, GLFW_KEY_W);
		addKey(Keysym.X, GLFW_KEY_X);
		addKey(Keysym.Y, GLFW_KEY_Z);
		addKey(Keysym.Z, GLFW_KEY_Y);
		
		addKey(Keysym.F1, GLFW_KEY_F1);
		addKey(Keysym.F2, GLFW_KEY_F2);
		addKey(Keysym.F3, GLFW_KEY_F3);
		addKey(Keysym.F4, GLFW_KEY_F4);
		addKey(Keysym.F5, GLFW_KEY_F5);
		addKey(Keysym.F6, GLFW_KEY_F6);
		addKey(Keysym.F7, GLFW_KEY_F7);
		addKey(Keysym.F8, GLFW_KEY_F8);
		addKey(Keysym.F9, GLFW_KEY_F9);
		addKey(Keysym.F10, GLFW_KEY_F10);
		addKey(Keysym.F11, GLFW_KEY_F11);
		addKey(Keysym.F12, GLFW_KEY_F12);
		
		addKey(Keysym.KP_0, GLFW_KEY_KP_0);
		addKey(Keysym.KP_1, GLFW_KEY_KP_1);
		addKey(Keysym.KP_2, GLFW_KEY_KP_2);
		addKey(Keysym.KP_3, GLFW_KEY_KP_3);
		addKey(Keysym.KP_4, GLFW_KEY_KP_4);
		addKey(Keysym.KP_5, GLFW_KEY_KP_5);
		addKey(Keysym.KP_6, GLFW_KEY_KP_6);
		addKey(Keysym.KP_7, GLFW_KEY_KP_7);
		addKey(Keysym.KP_8, GLFW_KEY_KP_8);
		addKey(Keysym.KP_9, GLFW_KEY_KP_9);
		
		addKey(Keysym.SPACE, GLFW_KEY_SPACE);
		addKey(Keysym.CIRCUMFLEX, GLFW_KEY_GRAVE_ACCENT);
		addKey(Keysym.SHARP_S, GLFW_KEY_MINUS);
		addKey(Keysym.ACUTE, GLFW_KEY_EQUAL);
		addKey(Keysym.UE, GLFW_KEY_LEFT_BRACKET);
		addKey(Keysym.PLUS, GLFW_KEY_RIGHT_BRACKET);
		addKey(Keysym.OE, GLFW_KEY_SEMICOLON);
		addKey(Keysym.AE, GLFW_KEY_APOSTROPHE);
		addKey(Keysym.NUMBER_SIGN, GLFW_KEY_BACKSLASH);
		addKey(Keysym.LESS_THAN, GLFW_KEY_WORLD_1);
		addKey(Keysym.COMMA, GLFW_KEY_COMMA);
		addKey(Keysym.PERIOD, GLFW_KEY_PERIOD);
		addKey(Keysym.HYPHEN, GLFW_KEY_SLASH);
		
		addKey(Keysym.KP_DECIMAL, GLFW_KEY_KP_DECIMAL);
		addKey(Keysym.KP_DIVIDE, GLFW_KEY_KP_DIVIDE);
		addKey(Keysym.KP_MULTIPLY, GLFW_KEY_KP_MULTIPLY);
		addKey(Keysym.KP_SUBTRACT, GLFW_KEY_KP_SUBTRACT);
		addKey(Keysym.KP_ADD, GLFW_KEY_KP_ADD);
		addKey(Keysym.KP_ENTER, GLFW_KEY_KP_ENTER);
		addKey(Keysym.KP_EQUAL, GLFW_KEY_KP_EQUAL);
		
		addKey(Keysym.CAPS_LOCK, GLFW_KEY_CAPS_LOCK);
		addKey(Keysym.SCROLL_LOCK, GLFW_KEY_SCROLL_LOCK);
		addKey(Keysym.NUM_LOCK, GLFW_KEY_NUM_LOCK);
		
		addKey(Keysym.ESCAPE, GLFW_KEY_ESCAPE);
		addKey(Keysym.RETURN, GLFW_KEY_ENTER);
		addKey(Keysym.TAB, GLFW_KEY_TAB);
		addKey(Keysym.BACKSPACE, GLFW_KEY_BACKSPACE);
		addKey(Keysym.INSERT, GLFW_KEY_INSERT);
		addKey(Keysym.DELETE, GLFW_KEY_DELETE);
		addKey(Keysym.RIGHT, GLFW_KEY_RIGHT);
		addKey(Keysym.LEFT, GLFW_KEY_LEFT);
		addKey(Keysym.DOWN, GLFW_KEY_DOWN);
		addKey(Keysym.UP, GLFW_KEY_UP);
		addKey(Keysym.PAGE_UP, GLFW_KEY_PAGE_UP);
		addKey(Keysym.PAGE_DOWN, GLFW_KEY_PAGE_DOWN);
		addKey(Keysym.HOME, GLFW_KEY_HOME);
		addKey(Keysym.END, GLFW_KEY_END);
		addKey(Keysym.PRINT_SCREEN, GLFW_KEY_PRINT_SCREEN);
		addKey(Keysym.PAUSE, GLFW_KEY_PAUSE);
		addKey(Keysym.LEFT_SHIFT, GLFW_KEY_LEFT_SHIFT);
		addKey(Keysym.LEFT_CONTROL, GLFW_KEY_LEFT_CONTROL);
		addKey(Keysym.ALT, GLFW_KEY_LEFT_ALT);
		addKey(Keysym.LEFT_SUPER, GLFW_KEY_LEFT_SUPER);
		addKey(Keysym.RIGHT_SHIFT, GLFW_KEY_RIGHT_SHIFT);
		addKey(Keysym.RIGHT_CONTROL, GLFW_KEY_RIGHT_CONTROL);
		addKey(Keysym.ALT_GRAPH, GLFW_KEY_RIGHT_ALT);
		addKey(Keysym.RIGHT_SUPER, GLFW_KEY_RIGHT_SUPER);
		addKey(Keysym.MENU, GLFW_KEY_MENU);
	}
	
	public static boolean createWindow(String id, boolean gldebug) {
		glfwDefaultWindowHints();
		glfwWindowHint(GLFW_VISIBLE, GLFW_FALSE);
		glfwWindowHint(GLFW_RESIZABLE, GLFW_TRUE);
		glfwWindowHint(GLFW_CONTEXT_DEBUG, gldebug ? GLFW_TRUE : GLFW_FALSE);
		window = glfwCreateWindow(100, 100, id, NULL, NULL);
		if(window == NULL) {
			Log.SYSTEM.error("Konnte kein Fenster für das Programm erstellen");
		}
		else {
			glfwSetMouseButtonCallback(window, new GLFWMouseButtonCallback() {
				public void invoke(long window, int button, int action, int mods) {
					if(action != GLFW_REPEAT && button >= 0 && button < Button.values().length)
						EVENTS.add(new WindowEvent(WindowAction.BUTTON, button, action == GLFW_PRESS ? 1 : 0));
				}
			});
			glfwSetCharCallback(window, new GLFWCharCallback() {
				public void invoke(long window, int codepoint) {
					EVENTS.add(new WindowEvent(WindowAction.CHARACTER, codepoint, 0));
				}
			});
			glfwSetCursorPosCallback(window, new GLFWCursorPosCallback() {
				public void invoke(long window, double xpos, double ypos) {
					EVENTS.add(new WindowEvent(WindowAction.CURSOR, (int)xpos, (int)ypos));
				}
			});
			glfwSetFramebufferSizeCallback(window, new GLFWFramebufferSizeCallback() {
				public void invoke(long window, int width, int height) {
					EVENTS.add(new WindowEvent(WindowAction.RESIZE, width, height));
				}
			});
			glfwSetKeyCallback(window, new GLFWKeyCallback() {
				public void invoke(long window, int key, int scancode, int action, int mods) {
					Keysym code = key > 0 && key <= KEYS.length ? KEYS[key - 1] : null;
					if(code != null)
						EVENTS.add(new WindowEvent(WindowAction.KEY, code.ordinal(), 
							(action == GLFW_PRESS ? KeyEvent.PRESS : (action == GLFW_RELEASE ? KeyEvent.RELEASE : KeyEvent.REPEAT)).ordinal()));
				}
			});
			glfwSetScrollCallback(window, new GLFWScrollCallback() {
				public void invoke(long window, double xoffset, double yoffset) {
					EVENTS.add(new WindowEvent(WindowAction.SCROLL, (int)xoffset, (int)yoffset));
				}
			});
			glfwSetWindowCloseCallback(window, new GLFWWindowCloseCallback() {
				public void invoke(long window) {
					EVENTS.add(new WindowEvent(WindowAction.CLOSED, 0, 0));
				}
			});
			glfwSetWindowFocusCallback(window, new GLFWWindowFocusCallback() {
				public void invoke(long window, boolean focused) {
					EVENTS.add(new WindowEvent(WindowAction.FOCUS, focused ? 1 : 0, 0));
				}
			});
//			glfwSetWindowSizeCallback(window, new GLFWWindowSizeCallback() {
//				public void invoke(long window, int width, int height) {
//					EVENTS.add(new WindowEvent(WindowAction.RESIZE, width, height));
//				}
//			});
			glfwSetWindowPosCallback(window, new GLFWWindowPosCallback() {
				public void invoke(long window, int xpos, int ypos) {
					EVENTS.add(new WindowEvent(WindowAction.POSITION, xpos, ypos));
				}
			});
			glfwSetWindowRefreshCallback(window, new GLFWWindowRefreshCallback() {
				public void invoke(long window) {
					EVENTS.add(new WindowEvent(WindowAction.REDRAW, 0, 0));
				}
			});
			glfwMakeContextCurrent(window);
			GL.createCapabilities();
		}
		return window != NULL;
	}
	public static void destroyWindow() {
		if(window != NULL) {
			glfwFreeCallbacks(window);
			glfwDestroyWindow(window);
			window = NULL;
		}
	}
	public static void initWindow(int sx, int sy, int wx, int wy) {
		if(window == NULL)
			return;
		long monitor = glfwGetPrimaryMonitor();
		GLFWVidMode mode = monitor == NULL ? null : glfwGetVideoMode(monitor);
		int[] x = new int[1];
		int[] y = new int[1];
		if(monitor != NULL)
			glfwGetMonitorPos(monitor, x, y);
		int xsize = (mode != null && wx > mode.width()) ? mode.width() : wx;
		int ysize = (mode != null && wy > mode.height()) ? mode.height() : wy;
		int xpos = x[0] + (mode != null ? mode.width() / 2 - xsize / 2 : 0);
		int ypos = y[0] + (mode != null ? mode.height() / 2 - ysize / 2 : 0);
		xsize = xsize < 1 ? 1 : xsize;
		ysize = ysize < 1 ? 1 : ysize;
		if(sx != 0x80000000 && sy != 0x80000000) {
			xpos = sx;
			ypos = sy;
		}
		if(wx != 0 && wy != 0) {
			xsize = wx;
			ysize = wy;
		}
		glfwSetWindowSize(window, xsize, ysize);
		glfwSetWindowPos(window, xpos, ypos);
//		wcf_windowed(win, xpos, ypos, xsize, ysize);
		// wcf_fullscreen does not move while hidden ...
//		wcf_setpos(win, xpos, ypos);
		glfwShowWindow(window);
//		wcf_show(win, 1);
//		wcf_limits(win, 1, 1, -1, -1);
	}
	
	public static WindowEvent[] poll() {
		glfwPollEvents();
		WindowEvent[] events = EVENTS.toArray(new WindowEvent[EVENTS.size()]);
		EVENTS.clear();
		return events;
	}
	public static DisplayMode[] getDisplayModes() {
		long monitor = glfwGetPrimaryMonitor();
		GLFWVidMode.Buffer data = monitor == NULL ? null : glfwGetVideoModes(monitor);
		if(data == null)
			return null;
		DisplayMode[] modes = new DisplayMode[data.limit()];
		for(int z = 0; z < data.limit(); z++) {
			GLFWVidMode mode = data.get(z);
			modes[z] = new DisplayMode(mode.width(), mode.height(), mode.refreshRate());
		}
		return modes;
	}
	public static DisplayMode getDisplayMode() {
		long monitor = glfwGetPrimaryMonitor();
		GLFWVidMode mode = monitor == NULL ? null : glfwGetVideoMode(monitor);
		if(mode == null)
			return null;
		return new DisplayMode(mode.width(), mode.height(), mode.refreshRate());
	}
	public static void init() {
		GLFWErrorCallback.create(new GLFWErrorCallback() {
            private Map<Integer, String> codes = APIUtil.apiClassTokens((field, value) -> 0x10000 < value && value < 0x20000, null, GLFW.class);
            
			public void invoke(int error, long description) {
				Log.SYSTEM.error("LWJGL-Fehler %s: %s", this.codes.get(error), getDescription(description));
			}
		}).set();
		if(!glfwInit())
			throw new IllegalStateException("Kann GLFW nicht initialisieren");
		if(glfwGetTimerFrequency() < 1000000L)
			throw new IllegalStateException("Timer unterstützt keine Mikrosekunden-Präzision");
		Log.SYSTEM.info("LWJGL " + Version.getVersion());
	}
	public static void end() {
		glfwTerminate();
		glfwSetErrorCallback(null).free();
	}
}
